#helper file for LSTM model evaluation

# helper functions for model evaluation and for metrics


#basic libraries
import numpy as np
import pandas as pd
import math

#LSTM:
import tensorflow as tf
from tensorflow import keras
from keras import layers

#Early stopping
from keras.callbacks import EarlyStopping

#Evaluation
from sklearn import metrics

def extractPredictions(data, step, start=0):
    return data.transpose()[start::step]


# Evaluates predictiosn againts dataset and returns requested metrics (rmse, mse, r2, mae)
def evaluatePredictions(predictions, dataset):
    r2 = []
    mae = []
    mse = []
    rmse = []
    ioa = [] #this is for index of agreement

    d = dataset.tail(len(predictions[0]))
    predictions[0].columns = [0]
    #var_r2 = np.var(d.iloc[:, 0])
    #print(var_r2)
    # Accomodate for data "Loss" for first few training values of test dataset
    for i in predictions:
        # Mogu da se dodaju dodatne metrike po potrebi
        r2.append(metrics.r2_score(d, i))
        mae.append(metrics.mean_absolute_error(d, i))
        mse.append(metrics.mean_squared_error(d, i))
        rmse.append(math.sqrt(metrics.mean_squared_error(d, i)))
        ioa.append(index_agreement(i.values,d.values))

        #r2.append(np.round(1-(mse[-1]/var_r2),5))
    #print("D:",d)
    #print("R2",r2)
    #print("MSE",mse)
    return r2, mae, mse, rmse, ioa


def evaluatePredictionsOverall(predictions, dataset):
    #TODO 
    overallR2 = 0
    overallMae = 0
    overallMse = 0
    overallRmse = 0
    overallIOA = 0 #this is for index of agreement
    d = dataset.tail(len(predictions[0])).to_numpy()

    overallMean = 0
    l=0
    for i in predictions:
        i.columns = [0]
        k = 0
        for j in i[0]:
                overallMean += (j)
                k+=1     
                l+=1
    overallMean = overallMean/l
    
    #Overall mean aposlute error:
    # sum(absolute(predictedValue - ActualValue))/Total number of values
    l=0
    for i in predictions:
        k = 0
        for j in i[0]:
                overallMae += abs((j - d[k]))
                k+=1
                l+=1
    overallMae = overallMae/l
    
    #Overall mean square error:
    # sum((predictedValue - ActualValue)^2)/Total number of values
    l=0
    for i in predictions:
        k = 0
        for j in i[0]:
                overallMse += ((j - d[k]))**2
                k+=1
                l+=1
    overallMse = overallMse/l
    
    #R2
    var = np.var(d)
    
    overallR2 = 1.0 - (overallMse/var)
    #Overall Root mean square error
    overallRmse = math.sqrt(overallMse)

    #this is now for index of agreement
    l = 0
    for i in predictions:
        overallIOA += index_agreement_overall(i, d)
        l += 1
    overallIOA = overallIOA / l
    #print(type(overallIOA))

    
    return overallR2[0], overallMae[0], overallMse[0], overallRmse, overallIOA[0]


# Relables result data
def LSTMpostProcessing(results, steps, inputs, test):
    results = pd.DataFrame(results)
    if (steps > 1):
        return results.set_index(test.index[(inputs):-(steps - 1)])
    return results.set_index(test.index[(inputs):])
    #return results.set_index(test.index[:results.shape[0]])


def extractResults(data, test, steps, inputs, start=0):
    results = []
    for i in range(steps):
        r = LSTMpostProcessing(pd.DataFrame(extractPredictions(data, steps, i)), steps, inputs, test)
        results.append(r)
    return results


#auxiliary functions for calculating the index of agreement (IOA)

def filter_nan(s, o):
    """
    this functions removed the data  from simulated and observed data
    whereever the observed data contains nan

    this is used by all other functions, otherwise they will produce nan as
    output
    """
    if np.sum(~np.isnan(s * o)) >= 1:
        data = np.array([s.flatten(), o.flatten()])
        data = np.transpose(data)
        data = data[~np.isnan(data).any(1)]
        s = data[:, 0]
        o = data[:, 1]
    return s, o

#index of agreement
def index_agreement(s, o):
    """
    index of agreement

    Willmott (1981, 1982)
    input:
        s: simulated
        o: observed
    output:
        ia: index of agreement
    """
    s,o = filter_nan(s,o)
    ia = 1.0 -(np.sum((o-s)**2))/(np.sum(
    (np.abs(s-np.mean(o))+np.abs(o-np.mean(o)))**2))
    return ia

def index_agreement_overall(s, o):
    """
    index of agreement

    Willmott (1981, 1982)
    input:
        s: simulated
        o: observed
    output:
        ia: index of agreement
    """
    ia = 1.0 - (np.sum((o-s)**2))/(np.sum(
    (np.abs(s-np.mean(o))+np.abs(o-np.mean(o)))**2))
    return ia

