'''

function for RNN attention model:
several stacked RNN layers are placed
and at the end, only the attention layer goes, after all the RNN layers


'''
# osnovne biblioteke
import numpy as np
import pandas as pd
import math
from attention import Attention

# set to run on CPU


# Early stopping
from keras.callbacks import EarlyStopping

# Evaluation
from sklearn import metrics

# import funkcije za evaluaciju modela
from utilities.LSTMEvaluate import *

# iz load data za scaling
from utilities.load_dataset_multi import *

import warnings

import os


class LSTMFunction:

    def __init__(self, trainData, valData, testData, test, test_denorm, D, features, lags, steps, intParams, bounds,
                 early_stop=False, target=None, scaler=None, norm_metrics=True):
        # time series, TimeSeries generator with train, wave and test data
        self.trainData = trainData
        self.valData = valData
        self.testData = testData

        # a test is also needed in the usual form for evaluation, this is a normalized test
        self.test = test

        # this is a denormalized test if we want denormalized metrics
        self.test_denorm = test_denorm

        # number of solution parameters
        self.D = D

        # this is for the scaler if we want the metrics as denormalized later
        # however, the scaler is loaded from load_datasetmulti and there is no need, so that's why
        # in constructor None
        self.scaler = scaler

        # whether we want to return normalized metrics or denormalized ones
        self.norm_metrics = norm_metrics

        # number of features, with univariate the number of features=1
        self.features = features

        # number of lags, how many lags we have, in the univariate code, the input of the network is
        # of shapes [1,number_of_lags], e.g. [1.6]
        self.lags = lags

        # number of steps forward, this will be the number of neurons in the outputlayer
        self.steps = steps

        # list with positions of metaheuristic parameters, where parameters are int
        self.intParams = intParams

        # dictionary with parameter value limits

        self.bounds = bounds

        # inicializing early_stop

        self.early_stop = early_stop

        # setting array with parameter value bounds
        # D is the total length of the string and is passed from the main code

        self.ub = [None] * self.D
        self.lb = [None] * self.D
        self.name = "RNN Attention Function"

        # we have only one parameter to start with, which is the number of neurons in the layer

        '''
        parameters:
        the idea is that real parameters go first, then integers, to make it easier to understand
        [0] - learning rate 
        [1] - dropout rate for the dropout layer that goes at the end
        [2]  - number of training epochs
        [3] - number of layers
        [4] - number of neurons in layers
        [5] - Same as number of neurons in layers 
        [6] - Same as number of neurons in layers 
        [7] - ... here it depends on how many parameters there are from the max number of layers. if max number of layers=n, then all from [4] onwards are the same boundaries
              for example, if max number of layers=2, then we have, in addition to the first 4 parameters ([0]-[3]), another 4+1 parameters (one for each number of neurons in each layer and one for the number of neurons in the attention layer) ,
              because we are building a network so that one lstm, one attention goes, etc.
        '''

        self.lb[0] = bounds['lb_lr']  # lower bound for learning rate, float
        self.ub[0] = bounds['ub_lr']  # lower bound for learning rate, float

        self.lb[1] = bounds['lb_dropout']  # lower bound for dropout
        self.ub[1] = bounds['ub_dropout']  # upper bound for dropout

        self.lb[2] = bounds['lb_epochs']  # lower bound for number of epochs, int
        self.ub[2] = bounds['ub_epochs']  # upper bound for number of epochs, int

        self.lb[3] = bounds['lb_layers']  # lower bound for number of layers, int
        self.ub[3] = bounds['ub_layers']  # upper bound for number of layers, int

        # now until the end of this sequence we set the value for the lower and upper limit of the number of neurons for all layers, including the attention layer
        # so that's why it goes to the end, and the length of D depends on ub_layers, so if ub_layers is e.g. 2, then the length of the string is D=4 (basic 4 parameters) + 2*2 = 8

        for i in range(4, D):
            self.lb[i] = bounds['lb_nn']  # lower bound for the number of neurons per layer, int
            self.ub[i] = bounds['ub_nn']  # upper bound for the number of neurons per layer, int, the last one is for the attention layer

        # the rest are fixed parameters for the start
        self.loss = 'mse'
        self.metrics = ['mean_absolute_error']
        self.epochs = 200
        self.droput = 0.01
        self.recurrent_droput = 0.01
        self.activation = 'relu'
        self.loss = 'mean_squared_error'
        self.learning_rate = 0.001

        # warnings.warn = self.warn

        # this is for target columnb for multivariate
        self.target = target

    # evaluation function
    def function(self, x):
        # sheet ace with the number of neurons
        # swarm individual is already a list and then you don't need to change anything
        # we will take only the first element of the list in case we have more parameters
        # print(type(x))
        # print(x)
        # we convert it back into a list, because it gets complicated with more hyperparameters

        learning_rate = x[0]
        dropout = x[1]  # this is for the dropout layer
        epochs = int(x[2])  # epoch is also an integer
        layers_lstm = int(x[3])   # number of layers
        # now we make a list of neurons, for each LSTM layer one neuron plus one at the end of the number of neurons for the attention layer
        # we go from the 4th index to the end as many layers as we have for LSTM, but we will skip the last one, since it is the last one for attention
        nn_list = x[4:(layers_lstm +4)]
        #convert everything to int
        nn_list = [int(x) for x in nn_list]
        #number of neurons for attention layer, no_attention is the first after the last number of neurons in the last lstm layer
        nn_attention = int(x[len(nn_list)+4])


        #print('NN list:',nn_list)
        #print('NN attention:',nn_attention)





        # LSTM model creation, training and prediction
        LSTMModel = self.createRNNAttentionModel(self.features, self.lags, self.steps, nn_list,nn_attention,
                                                   learning_rate=learning_rate)
        # self.trainLSTM(LSTMModel,self.epochs,self.trainData,self.valData)
        self.trainLSTM(LSTMModel, epochs, self.trainData, self.valData, earlyStop=self.early_stop)
        prediction = self.predictLSTM(LSTMModel, self.testData)

        # extract results and display
        results = extractResults(prediction, self.test, self.steps, self.lags)

        # print('RESULTS:',results)

        # this is a special step if we want to work with denormalized metrics

        dnormResults = denormalizeResultsEvaluation(results, self.test, self.target, self.steps, self.lags, self.test)

        # evaluatepredictions and evluatepredictionsoverall return in this order: r2, mae, mse, rmse
        # if there are more steps, the first element is r2 for all steps, then mae for all steps, etc.
        # for example, if we have 3 steps, then it is [r2first,r2second,r2third],[mafirst,masecond,masecond], etc.
        # these functions return tuples

        # here we first save all results, all metrics for all steps, if set to norm_metric1, then the main ones are normalized
        # and everything is generated and calculated according to the normalized ones
        # if self.norm_metrics=False, then the principals are denormalized and they are returned as primary

        if self.norm_metrics:  # we look at whether we want to return normalized or denormalized metrics
            allResults = evaluatePredictions(results, self.test[self.target])
            # here we now store the overall results, the average of all metrics for all steps
            overallResults = evaluatePredictionsOverall(results, self.test[self.target])

            allResults1 = evaluatePredictions(dnormResults, self.test_denorm[self.target])
            # print("ALL:",allResults)
            # here we now store the overall results, the average of all metrics for all steps
            overallResults1 = evaluatePredictionsOverall(dnormResults, self.test_denorm[self.target])

        else:
            allResults = evaluatePredictions(dnormResults, self.test_denorm[self.target])
            # print("ALL:",allResults)
            # here we now store the overall results, the average of all metrics for all steps
            overallResults = evaluatePredictionsOverall(dnormResults, self.test_denorm[self.target])

            allResults1 = evaluatePredictions(results, self.test[self.target])
            # here we now store the overall results, the average of all metrics for all steps
            overallResults1 = evaluatePredictionsOverall(results, self.test[self.target])

        # function returns the following:
        # therefore, we return two types of metrics, normalized and denormalized, and which will be primary depends on the flags self.normResults
        # objective, overall_results, all_results, overall_results1, all_results1, results, x (these are the solution parameters) and finally the model

        # first we define the objective, let the objective be the average mse (r2 should be larger and therefore cannot be used as an objective, when we do minimization problems)
        # objective = 1-overallResults[0]
        objective = overallResults[2]  # mse
        # objective = overallResults[2]
        #print(list(overallResults))
        # here, therefore, we return both normalized and denormalized metrics, and the main ones are overallResults and allResults, and calculations are made based on them
        return objective, list(overallResults), list(allResults), list(overallResults1), list(
            allResults1), results, LSTMModel

        # Help function for creating LSTM models - LSTMModel Creation
    def createRNNAttentionModel(self, feature_size, inputs, outputs, nn_list, nn_attention, dropout=0.01,
                                 recurrent_dropout=0.01,
                                 loss='mean_squared_error', optimizer='adam', activation='relu', metrics=['accuracy'],
                                 learning_rate=0.001, batch_size=32):

        optimizer = keras.optimizers.Adam(learning_rate=learning_rate)
        # optimzer = 'adam'

        LSTMModel = keras.models.Sequential()
        LSTMModel.add(keras.Input(shape=(inputs, feature_size)))  # Input_Size, sequence_length

        # now we go through the list of neurons
        for i in range(len(nn_list)):
            # samo u prvom LSTM layeru dodajemo batch_size
            if i==0:
               LSTMModel.add(layers.SimpleRNN(nn_list[i], activation=activation, batch_size=batch_size,return_sequences=True))
            else:
                LSTMModel.add(layers.SimpleRNN(nn_list[i], activation=activation,  return_sequences=True))
         #at the end, when we go through all the LSTM layers, we add attention
        LSTMModel.add(Attention(nn_attention))
        #we are adding a dropout layer
        LSTMModel.add(layers.Dropout(dropout))
        LSTMModel.add(layers.Dense(outputs))
        LSTMModel.compile(loss=loss, optimizer=optimizer, metrics=metrics)

        return LSTMModel

    def trainLSTM(self, model, epochs, trainData, valData=None, earlyStop=False):
        if (earlyStop):
            es = EarlyStopping(monitor='val_loss', mode='min', verbose=False, patience=epochs / 3)
            model.fit(trainData, epochs=epochs, verbose=False, validation_data=valData, callbacks=[es])
        else:
            if valData is None:
                model.fit(trainData, epochs=epochs, verbose=False)
            else:
                model.fit(trainData, epochs=epochs, verbose=False, validation_data=valData)

    def predictLSTM(self, model, testData):
        predictions = []
        predictions.append(model.predict(testData, verbose=False).reshape((-1)))
        predictions = np.array(predictions)
        return predictions

    # ignore warnings
    def warn(*args, **kwargs):
        pass

