# function to load time series for multivariate

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import math
# Timeseries sequence generator
from keras.preprocessing.sequence import TimeseriesGenerator

# VMD
from vmdpy import VMD
import PyEMD
pd.options.mode.chained_assignment = None



# Link: https://github.com/vrcarva/vmdpy


# Loads dataset, and indexes it based on time intervals

# Needs the time column to be provided in order to work:
# function for load dataset, returns a dataframe with the required column
# optionally, start and enddate can be passed, where start and enddate are indices
def loadTimeSeriesDataset(link, timedateColumn, columns=None, start=None, end=None):
    df = pd.read_csv(link)
    # df = df.dropna()
    # we drop on, but this drops all columns where we have at least one 0 value and that's why it doesn't work
    # df = df.dropna(inplace=True)
    # for utc=True add only if we have utc time format, so except date and hours and minutes
    datetime_series = pd.to_datetime(df[timedateColumn], utc=True)
    datetime_index = pd.DatetimeIndex(datetime_series.values)
    df = df.set_index(datetime_index)
    df = df.drop(timedateColumn, axis=1)
    if (columns != None):
        df = df[columns]
    # Subset from start to end
    if (start != None):
        df = df.loc[start:]
    if (end != None):
        df = df.loc[:end]
    df = df.dropna()  # this goes here because then it also cuts off the column that has all the values
    return df


# dataset plotting function
# shows the part of the dataset that is for train, validation and test
def plotSplit(dataset, train, test, val=None, path=None, dataset_name=None, y_label=None):
    plt.clf()
    plt.figure(figsize=(12, 4), dpi=300)
    #plt.style.use('seaborn')
    plt.style.use('fivethirtyeight')
    params = {'axes.labelsize': 14,
              'axes.titlesize': 14}
    plt.rcParams.update(params)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    if (val == None):
        dfTrain = dataset[0:int(len(dataset) * (train / 100))]
        dfTest = dataset[int(len(dataset) * (train / 100)):]
        plt.plot(dfTrain, label='Train')
        plt.plot(dfTest, label='Test')
    else:
        dfTrain = dataset[0:int(len(dataset) * (train / 100))]
        dfVal = dataset[
                int(len(dataset) * (train / 100)): int(len(dataset) * (train / 100)) + int(len(dataset) * (val / 100))]
        dfTest = dataset[int(len(dataset) * (train / 100)) + int(len(dataset) * (val / 100)):]
        plt.plot(dfTrain, label='Train')
        plt.plot(dfVal, label='Validation')
        plt.plot(dfTest, label='Test')
    plt.title(str(dataset_name) + ' dataset split')
    plt.ylabel(str(y_label))
    plt.xlabel('Date')
    plt.xticks(rotation=45)
    plt.legend(fontsize=14,bbox_to_anchor=(1.02, 1.02))
    plt.savefig(path + '/' + 'Dataset Visual.pdf', bbox_inches='tight')
    # plt.show()


# method for plotting series with VMD
# K = number of time series, so K+1 goes with the residual
# number of features is the real number of features
def plotSplitVMD(dataset, path=None, dataset_name=None, number_of_features=None, K=3, y_label=None):
    i = 0
    while i < len(dataset.columns):
        plt.clf()
        plt.figure(figsize=(12, 4), dpi=300)
        #plt.style.use('seaborn')
        plt.style.use('fivethirtyeight')
        params = {'axes.labelsize': 14,
                  'axes.titlesize': 14}
        plt.rcParams.update(params)
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        plt.xticks(rotation=45)
        j = 0
        while j < (K + 1):
            plt.plot(dataset.iloc[:, i + j], label=str(dataset.columns[i + j]))
            j = j + 1
        plt.title(
            str(dataset_name) + ' VMD decomposition of ' + str(dataset.columns[i + j-1][:dataset.columns[i].index('_')]))
        plt.ylabel(str(y_label))
        plt.xlabel('Date')
        plt.legend(fontsize=14,bbox_to_anchor=(1.02, 1.02))
        plt.savefig(path + '/' + 'VMD ' + str(dataset.columns[i] + '.pdf'), bbox_inches='tight')
        i = i + (K + 1)

#function for plotting EMD
#at EMD it goes to K+2 series, the last one is residual
def plotSplitEMD(dataset, path=None, dataset_name=None, number_of_features=None, K=3, y_label=None):
    i = 0
    while i < len(dataset.columns):
        plt.clf()
        plt.figure(figsize=(12, 4), dpi=300)
        #plt.style.use('seaborn')
        plt.style.use('fivethirtyeight')
        params = {'axes.labelsize': 14,
                  'axes.titlesize': 14}
        plt.rcParams.update(params)
        plt.xticks(fontsize=14)
        plt.yticks(fontsize=14)
        j = 0
        while j < (K + 2):
            plt.plot(dataset.iloc[:, i + j], label=str(dataset.columns[i + j]))
            j = j + 1
        plt.title(
            str(dataset_name) + ' EMD decomposition of ' + str(
                dataset.columns[i + j - 1][:dataset.columns[i].index('_')]))
        plt.ylabel(str(y_label))
        plt.xlabel('Date')
        plt.legend(fontsize=14, bbox_to_anchor=(1.02, 1.02))
        plt.savefig(path + '/' + 'EMD ' + str(dataset.columns[i] + '.pdf'), bbox_inches='tight')
        i = i + (K + 2)




from sklearn.preprocessing import MinMaxScaler

scaler = MinMaxScaler()


def normalizeData(Data):
    index = Data.index
    col = Data.columns
    scaler.fit(Data)
    Data = pd.DataFrame(scaler.transform(Data))
    Data = Data.set_index(index)
    Data.columns = col
    return Data


def denormalizeData(Data):
    index = Data.index
    columns = Data.columns
    Data = pd.DataFrame(scaler.inverse_transform(Data))
    Data = Data.set_index(index)
    Data.columns = columns
    return Data


def denormalizeResults(results, reference, target, lags):
    denormalizedResults = []
    for i in results:
        ref = reference.copy()
        ref[target] = i
        # ref = ref.fillna(0)
        ref = denormalizeData(ref)[lags:]
        # ref = denormalizeData(ref)
        ref.columns = reference.columns
        denormalizedResults.append(ref[target])
    return denormalizedResults




def denormalizeResultsEvaluation(results, reference, target, steps, inputs, test):
    denormalizedResults = []
    for i in results:
        ref = reference.copy()
        ref[target] = i
        ref = ref.fillna(0)
        if (steps > 1):
            ref = denormalizeData(ref)[inputs:-(steps - 1)]
        else:
            ref = denormalizeData(ref)[inputs:]
        ref.columns = test.columns
        denormalizedResults.append(ref[target])
    return denormalizedResults


# helpful to see something for statistics

def denormalizeResults1(results, reference, target, lags):
    denormalizedResults = []
    cols = [x for x in range(len(reference.columns))]
    for i in results:
        ref = reference.copy()
        ref[target] = i
        # ref = ref.fillna(0)
        ref = denormalizeData(ref)[lags:]
        # ref = denormalizeData(ref)
        ref.columns = reference.columns
        denormalizedResults.append(ref[target])
        denormalizedResults[cols].columns = cols
        cols = cols + 1
    return denormalizedResults


# Split data in to training nad testing or in to training, testing and validation data using paramaters:
# showPlot flag will display a plot of the data split
# Split data in to training nad testing or in to training, testing and validation data using paramaters:
# showPlot flag will display a plot of the data split
def datasetSplit(dataset, train, test, val=None, path=None):
    cols = dataset.columns;
    if (val == None):
        dfTrain = dataset[0:int(len(dataset) * (train / 100))]
        dfTrain.columns = cols
        dfTest = dataset[
                 int(len(dataset) * (train / 100)):]  # ovde staviti da ide do kraja, jer ima jedna manje obaservacija
        dfTest.columns = cols
        dfTest.to_csv(str(path) + '/' + 'test.csv')
        dfTrain.to_csv(str(path) + '/' + 'train.csv')
        return dfTrain, dfTest
    dfTrain = dataset[0:int(len(dataset) * (train / 100))]
    dfTrain.columns = cols
    dfVal = dataset[
            int(len(dataset) * (train / 100)): int(len(dataset) * (train / 100)) + int(len(dataset) * (val / 100))]
    dfVal.columns = cols
    # ovde staviti da ide do kraja, jer ima jedna manje obaservacija
    dfTest = dataset[int(len(dataset) * (train / 100)) + int(len(dataset) * (val / 100)):]
    dfTest.columns = cols
    dfTest.to_csv(str(path) + '/' + 'test.csv')
    dfTrain.to_csv(str(path) + '/' + 'train.csv')
    dfVal.to_csv(str(path) + '/' + 'val.csv')
    return dfTrain, dfVal, dfTest


# Time series generation

# Creates timeseries generators used to train and test a LSTM network

# Time series generation

# Creates timesearies generators used to train and test a LSTM network

def genTimeSeries(dataset, inputColumns, targetColumns, inputs, look_forward, stride=1, batch_size=None):
    if (batch_size == None):
        batch_size = dataset.shape[0]
    data_x = []
    data_y = []
    for i in range(len(dataset) - (look_forward - 1)):
        data_x.append(dataset[inputColumns].iloc[i])
    for i in range(len(dataset) - (look_forward - 1)):
        dyt = []
        for j in range(look_forward):
            dyt.append(dataset[targetColumns].iloc[i + j])
        data_y.append(dyt)
    return TimeseriesGenerator(np.array(data_x), np.array(data_y), length=inputs, stride=stride, batch_size=batch_size)



# some sample parameters for VMD
#alpha    moderate bandwidth constraint
#tau      noise-tolerance (no strict fidelity enforcement)
#K        Number of modes
#DC       Dc offset
#init     Omegas uniformly
#tol      Tolerance of convergence criterion
def applyVMDDecomposition(dataset, input_features, target, alpha=2000, tau=0., K=3, DC=0, init=1, tol=1e-7, path=None):
    new_features = []
    outDataSet = pd.DataFrame()

    if len(dataset) % 2 == 1:
        dataset = dataset[:-1]

    for t in target:
        outDataSet[t] = dataset[t]

    for i in input_features:
        u, u_hat, omega = VMD(dataset[i], alpha, tau, K, DC, init, tol)
        k = 0;
        for j in u:
            outDataSet[str(i) + '_VMD_Mode_' + str(k)] = j
            # Substracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_VMD_Mode_' + str(k))
            k += 1;
        # adds residual to dataset end
        outDataSet[str(i) + '_VMD_Residual'] = dataset[i]
        new_features.append(str(i) + '_VMD_Residual')
        outDataSet.to_csv(str(path) + '/' + 'Dataset after vmd.csv')
    return outDataSet, new_features


# Empirical mode decomposition (EMD)

# Controll pamaaters:
# IMFs = 10     #Max number of IMFs
# trials = 20   #Max number of trails
# method = 'akima'  other available method: 'piecewise cubic'

def applyEMDDecomposition(dataset, input_features, target, max_imf=10, max_trials=20, method='akima',path=None):
    new_features = []
    outDataSet = pd.DataFrame()

    emd = PyEMD.EMD(max_imf=max_imf, trials=max_trials, spline_kind=method)

    for t in target:
        outDataSet[t] = dataset[t]

    for i in input_features:
        IMFs = emd.emd(dataset[i].to_numpy(), max_imf=max_imf)
        k = 0;
        for j in IMFs:
            outDataSet[str(i) + '_EMD_IMF_' + str(k)] = j
            # Substracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_EMD_IMF_' + str(k))
            k += 1;
        # adds residual to dataset end
        outDataSet[str(i) + '_EMD_Residual'] = dataset[i]
        new_features.append(str(i) + '_EMD_Residual')
        outDataSet.to_csv(str(path) + '/' + 'Dataset after emd.csv')
    return outDataSet, new_features


# Ensemble Empirical mode decomposition (EEMD)

# Controll pamaaters:
# IMFs = 10     #Max number of decompositions
# trials = 20   #Max number of trails
# method = 'akima'  other available method: 'piecewise cubic'

def applyEEMDDecomposition(dataset, input_features, target, max_imf=10, max_trials=20, method='akima',path=None):
    new_features = []
    outDataSet = pd.DataFrame()

    emd = PyEMD.EEMD(max_imfs=max_imf, trials=max_trials, spline_kind=method)

    for t in target:
        outDataSet[t] = dataset[t]

    for i in input_features:
        IMFs = emd.eemd(dataset[i].to_numpy(), None, max_imf=max_imf)
        k = 0;
        for j in IMFs:
            outDataSet[str(i) + '_EEMD_IMF_' + str(k)] = j
            # Sutracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_EEMD_IMF_' + str(k))
            k += 1;
        # adds residual to dataset end
        outDataSet[str(i) + '_EEMD_Residual'] = dataset[i]
        new_features.append(str(i) + '_EEMD_Residual')
        outDataSet.to_csv(str(path) + '/' + 'Dataset after eemd.csv')
    return outDataSet, new_features


# Complete Ensemble Empirical Mode Decomposition with Adaptive Noise (CEEMDAN)

# Controll pamaaters:
# IMFs = 10     #Max number of decompositions
# trials = 20   #Max number of trials
# method = 'akima'  other available method: 'piecewise cubic'

def applyCEEMDANDecomposition(dataset, input_features, target, max_imf=10, max_trials=20, method='akima',path=None):
    new_features = []
    outDataSet = pd.DataFrame()

    emd = PyEMD.CEEMDAN(max_imfs=max_imf, trials=max_trials, spline_kind=method)

    for t in target:
        outDataSet[t] = dataset[t]

    for i in input_features:
        IMFs = emd.ceemdan(dataset[i].to_numpy(), None, max_imf=max_imf)
        k = 0;
        for j in IMFs:
            outDataSet[str(i) + '_CEEMDAN_IMF_' + str(k)] = j
            # Sutracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_CEEMDAN_IMF_' + str(k))
            k += 1;
        # adds residual to dataset end
        outDataSet[str(i) + '_CEEMDAN_Residual'] = dataset[i]
        new_features.append(str(i) + '_CEEMDAN_Residual')
        outDataSet.to_csv(str(path) + '/' + 'Dataset after ceemdan.csv')
    return outDataSet, new_features


# VMD + EEMD

# Control parameters:

# VMD:
# alpha    moderate bandwidth constraint
# tau      noise-tolerance (no strict fidelity enforcement)
# K        Number of modes
# DC       Dc offset
# init     Omegas uniformly
# tol      Tolerance of convergence criterion

# EEMD:
# IMFs = 10     #Max number of decompositions
# trials = 20   #Max number of trials
# method = 'akima'  other available method: 'piecewise cubic'

def applyVMDEEMDDecomposition(dataset, input_features, target,
                              # VMD PARAMETERS:
                              alpha=2000, tau=0., K=3, DC=0, init=1, tol=1e-7,
                              # EEMD PARAMETERS
                              max_imf=10, max_trials=20, method='akima',path=None):
    new_features = []
    outDataSet = pd.DataFrame()

    emd = PyEMD.EEMD(max_imfs=max_imf, trials=max_trials, spline_kind=method)

    if len(dataset) % 2 == 1:
        dataset = dataset[:-1]

    for t in target:
        outDataSet[t] = dataset[t]

    for i in input_features:
        u, u_hat, omega = VMD(dataset[i], alpha, tau, K, DC, init, tol)
        k = 0;
        for j in u:
            outDataSet[str(i) + '_VMD_Mode_' + str(k)] = j
            # Sutracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_VMD_Mode_' + str(k))
            k += 1;
        # Process VMD Residuals with EEMD residual to dataset end
        IMFs = emd.emd(dataset[i].to_numpy(), None, max_imf=max_imf)
        k = 0;
        for j in IMFs:
            outDataSet[str(i) + '_VMD_Residual_EEMD_IMF_' + str(k)] = j
            # Sutracts currect mode from total value(calculates residual iterativley)
            dataset[i] = dataset[i] - j
            new_features.append(str(i) + '_VMD_Residual_EEMD_IMF_' + str(k))
            k += 1;
        # adds residual to dataset end
        outDataSet[str(i) + '_VMD_Residual_EEMD_Residual'] = dataset[i]
        new_features.append(str(i) + '_VMD_Residual_EEMD_Residual')
        outDataSet.to_csv(str(path) + '/' + 'Dataset after vmd eemd.csv')
    return outDataSet, new_features